/*  OpenTally: Open-source election vote counting
 *  Copyright © 2021 Lee Yingtong Li (RunasSudo)
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

use opentally::election::{CandidateState, CountState, Election};
use opentally::numbers::{Number, Rational};
use opentally::stv;

use csv::StringRecord;

use std::fs::File;
use std::io::{self, BufRead}; 
use std::ops;

#[allow(dead_code)] // Suppress false positive
pub fn read_validate_election<N: Number>(csv_file: &str, blt_file: &str, stv_opts: stv::STVOptions) {
	// Read CSV file
	let reader = csv::ReaderBuilder::new()
		.has_headers(false)
		.from_path(csv_file)
		.expect("IO Error");
	let records: Vec<StringRecord> = reader.into_records().map(|r| r.expect("Syntax Error")).collect();
	
	let mut candidates: Vec<&str> = records.iter().skip(2).map(|r| &r[0]).collect();
	// Remove exhausted/LBF rows
	candidates.truncate(candidates.len() - 2);
	
	// TODO: Validate candidate names
	
	let stages: Vec<usize> = records.first().unwrap().iter().skip(1).step_by(2).map(|s| s.parse().unwrap()).collect();
	
	// Decompress BLT
	let file = File::open(blt_file).expect("IO Error");
	let file_reader = io::BufReader::new(file);
	let lines = file_reader.lines();
	
	// Read BLT
	let election: Election<Rational> = Election::from_blt(lines.map(|r| r.expect("IO Error").to_string()).into_iter());
	
	validate_election(stages, records, election, stv_opts);
}

pub fn validate_election<N: Number>(stages: Vec<usize>, records: Vec<StringRecord>, election: Election<N>, stv_opts: stv::STVOptions)
where
	for<'r> &'r N: ops::Sub<&'r N, Output=N>,
	for<'r> &'r N: ops::Div<&'r N, Output=N>,
	for<'r> &'r N: ops::Neg<Output=N>,
{
	// Initialise count state
	let mut state = CountState::new(&election);
	
	// Distribute first preferences
	stv::count_init(&mut state, &stv_opts);
	let mut stage_num = 1;
	
	for (idx, stage) in stages.into_iter().enumerate() {
		while stage_num < stage {
			// Step through stages
			// Assert count not yet done
			assert_eq!(stv::count_one_stage(&mut state, &stv_opts), false);
			stage_num += 1;
		}
		validate_stage(idx, &state, &records);
	}
}

fn validate_stage<N: Number>(idx: usize, state: &CountState<N>, records: &Vec<StringRecord>) {
	println!("Col at idx {}", idx);
	
	let mut candidate_votes: Vec<N> = records.iter().skip(2).map(|r| N::from(r[idx*2 + 1].parse::<f64>().expect("Syntax Error"))).collect();
	
	// Validate exhausted/LBF
	let lbf_votes = candidate_votes.pop().unwrap();
	let exhausted_votes = candidate_votes.pop().unwrap();
	
	assert!(state.exhausted.votes == exhausted_votes);
	assert!(state.loss_fraction.votes == lbf_votes);
	
	// Remove exhausted/LBF rows
	candidate_votes.truncate(candidate_votes.len() - 2);
	
	// Validate candidate votes
	for (candidate, votes) in state.election.candidates.iter().zip(candidate_votes) {
		let count_card = state.candidates.get(candidate).unwrap();
		assert!(count_card.votes == votes, "Failed to validate votes for candidate {}. Expected {:}, got {:}", candidate.name, votes, count_card.votes);
	}
	
	// Validate candidate states
	let mut candidate_states: Vec<&str> = records.iter().skip(2).map(|r| &r[idx*2 + 2]).collect();
	candidate_states.truncate(candidate_states.len() - 2);
	
	for (candidate, candidate_state) in state.election.candidates.iter().zip(candidate_states) {
		let count_card = state.candidates.get(candidate).unwrap();
		if candidate_state == "" {
			assert!(count_card.state == CandidateState::HOPEFUL);
		} else if candidate_state == "EL" || candidate_state == "PEL" {
			assert!(count_card.state == CandidateState::ELECTED);
		} else if candidate_state == "EX" || candidate_state == "EXCLUDING" {
			assert!(count_card.state == CandidateState::EXCLUDED);
		} else {
			panic!("Unknown state descriptor {}", candidate_state);
		}
	}
}
